@@ -19,9 +19,9 @@
 ;; compact needs to deal with it ...
 
 (define-module (nyacc lalr)
-  #:export-syntax (lalr-spec)
+  ;;#:export-syntax (lalr-spec)
   #:export (*nyacc-version*
-	    make-lalr-machine compact-machine hashify-machine 
+	    lalr-spec make-lalr-machine compact-machine hashify-machine 
 	    lalr-start lalr-match-table
 	    restart-spec add-recovery-logic!
 	    pp-lalr-notice pp-lalr-grammar pp-lalr-machine
@@ -42,7 +42,7 @@
   #:use-module (nyacc util)
   )
 
-(define *nyacc-version* "0.80.4")
+(define *nyacc-version* "0.81.0")
 
 ;; @deffn proxy-? sym rhs
 ;; @example
