@@ -1,19 +1,19 @@
-;;; lang/c99/body.scm
-;;;
-;;; Copyright (C) 2015-2017 Matthew R. Wette
-;;;
-;;; This program is free software: you can redistribute it and/or modify
-;;; it under the terms of the GNU General Public License as published by 
-;;; the Free Software Foundation, either version 3 of the License, or 
-;;; (at your option) any later version.
-;;;
-;;; This program is distributed in the hope that it will be useful,
-;;; but WITHOUT ANY WARRANTY; without even the implied warranty of 
-;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-;;; GNU General Public License for more details.
-;;;
-;;; You should have received a copy of the GNU General Public License
-;;; along with this program.  If not, see <http://www.gnu.org/licenses/>.
+;; lang/c99/body.scm
+;;
+;; Copyright (C) 2015-2017 Matthew R. Wette
+;;
+;; This program is free software: you can redistribute it and/or modify
+;; it under the terms of the GNU General Public License as published by 
+;; the Free Software Foundation, either version 3 of the License, or 
+;; (at your option) any later version.
+;;
+;; This program is distributed in the hope that it will be useful,
+;; but WITHOUT ANY WARRANTY; without even the implied warranty of 
+;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+;; GNU General Public License for more details.
+;;
+;; You should have received a copy of the GNU General Public License
+;; along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
 ;; Notes on the code design may be found in doc/nyacc/lang/c99-hg.info
 
@@ -71,6 +71,8 @@
       (cons (substring defstr 0 x3) (substring defstr (1+ x3)))))))
 
 ;; @deffn Procedure make-cpi debug defines incdirs inchelp
+;; I think there is a potential bug here in that the alist of cpp-defs/helpers
+;; should be last-in-first-seen ordered.  Probably helpers low prio.
 ;; @end deffn
 (define (make-cpi debug defines incdirs inchelp)
   ;; convert inchelp into inc-file->typenames and inc-file->defines
@@ -108,6 +110,8 @@
 	   (lambda (tl) (set-cpi-ctl! cpi (append tl (cpi-ctl cpi)))))
     (and=> (assoc-ref (cpi-idefd cpi) "__builtin")
 	   (lambda (tl) (set-cpi-defs! cpi (append tl (cpi-defs cpi)))))
+    ;;(sferr "cpi-defs=\n")
+    ;;(pretty-print (cpi-defs cpi) (current-error-port))
     ;; Return the populated info.
     cpi))
 
@@ -164,7 +168,7 @@
 ;; @end deffn
 (define (find-new-typenames decl)
 
-  ;; like declr->ident in util2.scm
+  ;; like declr-id in util2.scm
   (define (declr->id-name declr)
     (case (car declr)
       ((ident) (sx-ref declr 1))
@@ -216,7 +220,7 @@
 	    (if (eq? c2 #\newline)
 		(iter cl (read-char))
 		(iter (cons* c2 ch cl) (read-char)))))
-	 ((eq? ch #\/) ;; swallow comments, event w/ newlines
+	 ((eq? ch #\/) ;; swallow comments, even w/ newlines
 	  (let ((c2 (read-char)))
 	    (cond
 	     ((eqv? c2 #\*)
@@ -233,14 +237,6 @@
 	      (iter (cons #\/ cl) c2)))))
 	 (else (iter (cons ch cl) (read-char)))))))
 
-;; @deffn {Procedure} find-file-in-dirl file dirl => path
-;; @end deffn
-(define (find-file-in-dirl file dirl)
-  (let iter ((dirl dirl))
-    (if (null? dirl) #f
-	(let ((p (string-append (car dirl) "/" file)))
-	  (if (access? p R_OK) p (iter (cdr dirl)))))))
-
 (define (def-xdef? name mode)
   (not (eqv? mode 'file)))
 
@@ -259,17 +255,22 @@
 ;; @end example
 ;; @end deffn
 (define gen-c-lexer
-  ;; This gets ugly in order to handle cpp.
-  ;;.need to add support for num's w/ letters like @code{14L} and @code{1.3f}.
+  ;; This gets ugly in order to handle cpp.  The CPP will tokenize, expand,
+  ;; then convert back to a string.
+  ;;
+  ;; todo: check if @code{1.3f} gets parsed as a number.
   ;; todo: I think there is a bug wrt the comment reader because // ... \n
   ;; will end up in same mode...  so after
   ;; int x; // comment
   ;; the lexer will think we are not at BOL.
+  ;;
+  ;; The state variable `suppress' is used to suppress re-expansion of input
+  ;; text generated by the CPP macro expander.  The CPP replacement text
+  ;; inserted via a string-port on the port stack.  When that port is fully
+  ;; read (i.e., the reader sees eof-object) then @var{suppress} is changed
+  ;; to @code{#t}.
   (let* ((match-table c99-mtab)
-	 (read-ident read-c-ident)
-	 (read-comm read-c-comm)
-	 ;;
-	 (ident-like? (make-ident-like-p read-ident))
+	 (ident-like? (make-ident-like-p read-c-ident))
 	 ;;
 	 (strtab (filter-mt string? match-table)) ; strings in grammar
 	 (kwstab (filter-mt ident-like? strtab))  ; keyword strings =>
@@ -307,37 +308,34 @@
 	      (set-cpi-defs! info (cons cell (cpi-defs info)))))
 	  
 	  (define (rem-define name)
-	      (set-cpi-defs! info (acons name #f (cpi-defs info))))
+	    (set-cpi-defs! info (acons name #f (cpi-defs info))))
 	  
 	  (define (apply-helper file)
+	    ;;(when (string=? file "stddef.h") (pretty-print (cpi-itynd info)))
 	    (let* ((tyns (assoc-ref (cpi-itynd info) file))
 		   (defs (assoc-ref (cpi-idefd info) file)))
-	      ;;(simple-format #t "apply-helper ~S => ~S\n" file tyns)	 
-	      ;;(simple-format #t "  itynd= ~S\n" (cpi-itynd info))
 	      (when tyns
 		(for-each add-typename tyns)
 		(set-cpi-defs! info (append defs (cpi-defs info))))
 	      tyns))
-	  
-	  (define (inc-stmt->file stmt)
-	    (let* ((arg (cadr stmt)) (len (string-length arg)))
-	      (substring arg 1 (1- len))))
 
-	  (define (inc-file->path file)
-	    (find-file-in-dirl file (cpi-incs info)))
+	  (define (inc-stmt->file stmt) ;; retain <> or ""
+	    (let* ((arg (cadr stmt)))
+	      (if (ident-like? arg) ;; #include MYFILE
+		  (expand-cpp-macro-ref arg (cpi-defs info))
+		  arg)))
+
+	  (define (inc-file->path file next)
+	    (find-incl-in-dirl file (cpi-incs info) next))
 
 	  (define (code-if stmt)
-	    ;;(simple-format #t "code-if: ppxs=~S\n" ppxs)
 	    (case (car ppxs)
 	      ((skip-look skip-done skip) ;; don't eval if excluded
-	       ;;(simple-format #t "code-if: SKIP\n")
 	       (set! ppxs (cons 'skip ppxs)))
 	      (else
-	       ;;(simple-format #t "code-if: KEEP\n")
-	       ;;(simple-format #t "  text=~S\n" (cadr stmt))
-	       ;;(simple-format #t "  defs=~S\n" (cpi-defs info))
 	       (let* ((defs (cpi-defs info))
-		      (val (eval-cpp-cond-text (cadr stmt) defs)))
+		      (val (eval-cpp-cond-text (cadr stmt) defs
+					       #:inc-dirs (cpi-incs info))))
 		 (if (not val) (c99-err "unresolved: ~S" (cadr stmt)))
 		 (if (eq? 'keep (car ppxs))
 		     (if (zero? val)
@@ -351,7 +349,8 @@
 	      ((skip) #t) ;; don't eval if excluded
 	      (else
 	       (let* ((defs (cpi-defs info))
-		      (val (eval-cpp-cond-text (cadr stmt) defs)))
+		      (val (eval-cpp-cond-text (cadr stmt) defs
+					       #:inc-dirs (cpi-incs info))))
 		 (if (not val) (c99-err "unresolved: ~S" (cadr stmt)))
 		 (case (car ppxs)
 		   ((skip-look) (if (not (zero? val)) (set-car! ppxs 'keep)))
@@ -368,26 +367,28 @@
 	    (set! ppxs (cdr ppxs))
 	    stmt)
 	  
-	  (define (eval-cpp-incl/here stmt) ;; => stmt
+	  (define* (eval-cpp-incl/here stmt #:optional next) ;; => stmt
 	    (let* ((file (inc-stmt->file stmt))
-		   (path (inc-file->path file)))
+		   (path (inc-file->path file next)))
+	      ;;(sferr "include ~S\n" path)
 	      (cond
 	       ((apply-helper file))
 	       ((not path) (c99-err "not found: ~S" file)) ; file not found
 	       (else (set! bol #t) (push-input (open-input-file path))))
-	      stmt))
+	      (sx+attr* stmt 'path path)))
 
-	  (define (eval-cpp-incl/tree stmt) ;; => stmt
+	  (define* (eval-cpp-incl/tree stmt #:optional next) ;; => stmt
 	    ;; include file as a new tree
 	    (let* ((file (inc-stmt->file stmt))
-		   (path (inc-file->path file)))
+		   (path (inc-file->path file next)))
+	      ;;(sferr "include ~S\n" path)
 	      (cond
 	       ((apply-helper file) stmt)		 ; use helper
 	       ((not path) (c99-err "not found: ~S" file)) ; file not found
 	       ((with-input-from-file path run-parse) => ; add tree to stmt
 		(lambda (tree)
 		  (for-each add-define (xp1 tree))
-		  (append stmt (list tree)))))))
+		  (append (sx+attr* stmt 'path path) (list tree)))))))
 
 	  (define (eval-cpp-stmt/code stmt) ;; => stmt
 	    (case (car stmt)
@@ -399,11 +400,16 @@
 	       (if (eqv? 'keep (car ppxs))
 		   (case (car stmt)
 		     ((include) (eval-cpp-incl/here stmt))
+		     ((include-next) (eval-cpp-incl/here stmt 'next))
 		     ((define) (add-define stmt) stmt)
 		     ((undef) (rem-define (cadr stmt)) stmt)
 		     ((error) (c99-err "error: #error ~A" (cadr stmt)))
+		     ((warning) (report-error "warning: ~A" (cdr stmt)))
 		     ((pragma) stmt) ;; ignore for now
-		     (else (error "bad cpp flow stmt")))))))
+		     ((line) stmt)
+		     (else
+		      (sferr "stmt: ~S\n" stmt)
+		      (error "1: bad cpp flow stmt")))))))
 	       
 	  (define (eval-cpp-stmt/decl stmt) ;; => stmt
 	    (case (car stmt)
@@ -418,11 +424,19 @@
 		      (if (cpi-top-blev? info)
 			  (eval-cpp-incl/tree stmt)
 			  (eval-cpp-incl/here stmt)))
+		     ((include-next)	; gcc extension
+		      (if (cpi-top-blev? info)
+			  (eval-cpp-incl/tree stmt 'next)
+			  (eval-cpp-incl/here stmt 'next)))
 		     ((define) (add-define stmt) stmt)
 		     ((undef) (rem-define (cadr stmt)) stmt)
 		     ((error) (c99-err "error: #error ~A" (cadr stmt)))
+		     ((warning) (report-error "warning: ~A" (cdr stmt)) stmt)
 		     ((pragma) stmt) ;; ignore for now
-		     (else (error "bad cpp flow stmt")))
+		     ((line) stmt)
+		     (else
+		      (sferr "stmt: ~S\n" stmt)
+		      (error "2: bad cpp flow stmt")))
 		   stmt))))
 	       
 	  (define (eval-cpp-stmt/file stmt) ;; => stmt
@@ -434,8 +448,12 @@
 	      ((define) (add-define stmt) stmt)
 	      ((undef) (rem-define (cadr stmt)) stmt)
 	      ((error) stmt)
+	      ((warning) stmt)
 	      ((pragma) stmt) ;; need to work this
-	      (else (error "bad cpp flow stmt"))))
+	      ((line) stmt)
+	      (else
+	       (sferr "stmt: ~S\n" stmt)
+	       (error "3: bad cpp flow stmt"))))
 
 	  ;; Maybe evaluate the CPP statement.
 	  (define (eval-cpp-stmt stmt)
@@ -461,38 +479,27 @@
 	    (case mode
 	      ((code) #f)
 	      ((decl) (and (cpi-top-blev? info)
-			   (memq (car stmt) '(include define))))
+			   (memq (car stmt) '(include define include-next))))
 	      ((file) (or (cpi-top-blev? info)
-			  (not (eqv? (car stmt) 'include))))
+			  (not (memq (car stmt) '(include include-next)))))
 	      (else (error "lang/c99 coding error"))))
 
 	  ;; Composition of @code{read-cpp-line} and @code{eval-cpp-line}.
 	  (define (read-cpp-stmt ch)
 	    (and=> (read-cpp-line ch) cpp-line->stmt))
 
-	  ;; Recheck for expansions that turn from plain text to macro call:
-	  ;; @example
-	  ;; #define ABC(X,Y) ((X)+(Y))
-	  ;; #define DEF ABC
-	  ;; int x = DEF(1,2);
-	  ;; @end example
-	  (define (recheck-repl repl defs name)
-	    (if (ident-like? repl)
-		(or (expand-cpp-macro-ref repl defs (list name)) repl)
-		repl))
-
 	  (define (read-token)
 	    (let iter ((ch (read-char)))
 	      (cond
 	       ((eof-object? ch)
 		(set! suppress #f)
-		(if (pop-input) (iter (read-char)) (assc-$ '($end . ""))))
+		(if (pop-input) (iter (read-char)) (assc-$ '($end . "#<eof>"))))
 	       ((eq? ch #\newline) (set! bol #t) (iter (read-char)))
 	       ((char-set-contains? c:ws ch) (iter (read-char)))
 	       (bol
 		(set! bol #f)
 		(cond ;; things that depend on bol only
- 		 ((read-comm ch #t) => assc-$)
+ 		 ((read-c-comm ch #t #:skip-prefix #t) => assc-$)
 		 ((read-cpp-stmt ch) =>
 		  (lambda (stmt)
 		    (let ((stmt (eval-cpp-stmt stmt))) ; eval can add tree
@@ -500,7 +507,7 @@
 			  (assc-$ `(cpp-stmt . ,stmt))
 			  (iter (read-char))))))
 		 (else (iter ch))))
-	       ((read-ident ch) =>
+	       ((read-c-ident ch) =>
 		(lambda (name)
 		  (let ((symb (string->symbol name))
 			(defs (cpi-defs info)))
@@ -509,10 +516,9 @@
 			   (if (procedure? x-def?) (x-def? name mode) x-def?)
 			   (expand-cpp-macro-ref name defs))
 		      => (lambda (repl)
+			   ;;(sferr "repl=~S\n" repl)
 			   (set! suppress #t) ; don't rescan
-			   (push-input
-			    (open-input-string
-			     (recheck-repl repl defs name)))
+			   (push-input (open-input-string repl))
 			   (iter (read-char))))
 		     ((assq-ref keytab symb)
 		      => (lambda (t) (cons t name)))
@@ -523,7 +529,7 @@
 	       ((read-c-num ch) => assc-$)
 	       ((read-c-string ch) => assc-$)
 	       ((read-c-chlit ch) => assc-$)
-	       ((read-comm ch #f) => assc-$)
+	       ((read-c-comm ch #f #:skip-prefix #t) => assc-$)
 	       ((and (char=? ch #\{)	; ugly tracking of block-lev in lexer
 		     (eqv? 'keep (car ppxs)) (cpi-inc-blev! info) #f) #f)
 	       ((and (char=? ch #\})	; ugly tracking of block-lev in lexer
@@ -538,8 +544,9 @@
 
 	  ;; Loop between reading tokens and skipping tokens via CPP logic.
 	  (let iter ((pair (read-token)))
+	    ;;(report-error "lx iter=>~S" (list pair))
 	    (case (car ppxs)
-	      ((keep) ;;(simple-format #t "lx=>~S\n" pair)
+	      ((keep)
 	       pair)
 	      ((skip-done skip-look skip)
 	       (iter (read-token)))
