@@ -18,6 +18,9 @@
    (lambda ($1 . $rest) `(p-expr ,(tl->list $1)))
    ;; primary-expression => "(" expression ")"
    (lambda ($3 $2 $1 . $rest) $2)
+   ;; primary-expression => "(" "{" block-item-list "}" ")"
+   (lambda ($5 $4 $3 $2 $1 . $rest)
+     `(stmt-expr (@ (extension "GNUC")) ,$3))
    ;; postfix-expression => primary-expression
    (lambda ($1 . $rest) $1)
    ;; postfix-expression => postfix-expression "[" expression "]"
@@ -223,8 +226,13 @@
    (lambda ($3 $2 $1 . $rest) (tl-append $1 $3))
    ;; init-declarator => declarator
    (lambda ($1 . $rest) `(init-declr ,$1))
+   ;; init-declarator => declarator asm-expression
+   (lambda ($2 $1 . $rest) `(init-declr ,$1))
    ;; init-declarator => declarator "=" initializer
    (lambda ($3 $2 $1 . $rest) `(init-declr ,$1 ,$3))
+   ;; init-declarator => declarator asm-expression "=" initializer
+   (lambda ($4 $3 $2 $1 . $rest)
+     `(init-declr ,$1 ,$4))
    ;; storage-class-specifier => "auto"
    (lambda ($1 . $rest) '(stor-spec (auto)))
    ;; storage-class-specifier => "extern"
@@ -478,19 +486,19 @@
    ;; direct-declarator => direct-declarator "(" ")"
    (lambda ($3 $2 $1 . $rest)
      `(ftn-declr ,$1 (param-list)))
-   ;; pointer => "*" type-qualifier-list
-   (lambda ($2 $1 . $rest)
-     `(pointer ,(tl->list $2)))
-   ;; pointer => "*"
-   (lambda ($1 . $rest) '(pointer))
    ;; pointer => "*" type-qualifier-list pointer
    (lambda ($3 $2 $1 . $rest)
      `(pointer ,(tl->list $2) ,$3))
+   ;; pointer => "*" type-qualifier-list
+   (lambda ($2 $1 . $rest)
+     `(pointer ,(tl->list $2)))
    ;; pointer => "*" pointer
    (lambda ($2 $1 . $rest) `(pointer ,$2))
+   ;; pointer => "*"
+   (lambda ($1 . $rest) '(pointer))
    ;; type-qualifier-list => type-qualifier
    (lambda ($1 . $rest)
-     (make-tl 'decl-spec-list $1))
+     (make-tl 'type-qual-list $1))
    ;; type-qualifier-list => type-qualifier-list type-qualifier
    (lambda ($2 $1 . $rest) (tl-append $1 $2))
    ;; parameter-type-list => parameter-list
@@ -520,17 +528,17 @@
      `(type-name ,(tl->list $1) ,$2))
    ;; type-name => declaration-specifiers
    (lambda ($1 . $rest) `(type-name ,(tl->list $1)))
-   ;; abstract-declarator => pointer
-   (lambda ($1 . $rest) `(abs-declr ,$1))
    ;; abstract-declarator => pointer direct-abstract-declarator
    (lambda ($2 $1 . $rest) `(abs-declr ,$1 ,$2))
+   ;; abstract-declarator => pointer
+   (lambda ($1 . $rest) `(abs-declr ,$1))
    ;; abstract-declarator => direct-abstract-declarator
    (lambda ($1 . $rest) `(abs-declr ,$1))
    ;; direct-abstract-declarator => "(" abstract-declarator ")"
    (lambda ($3 $2 $1 . $rest) `(declr-scope ,$2))
    ;; direct-abstract-declarator => direct-abstract-declarator "[" type-qua...
    (lambda ($5 $4 $3 $2 $1 . $rest)
-     `(declr-array ,$3 ,$4))
+     `(declr-array ,$1 ,$3 ,$4))
    ;; direct-abstract-declarator => direct-abstract-declarator "[" type-qua...
    (lambda ($4 $3 $2 $1 . $rest)
      `(declr-array ,$1 ,$3))
@@ -582,19 +590,19 @@
         ,(tl->list (tl-insert '(stor-spec "static") $2))
         ,$4))
    ;; direct-abstract-declarator => direct-abstract-declarator "[" "*" "]"
-   (lambda ($4 $3 $2 $1 . $rest) `(declr-STAR ,$1))
+   (lambda ($4 $3 $2 $1 . $rest) `(declr-star ,$1))
    ;; direct-abstract-declarator => "[" "*" "]"
-   (lambda ($3 $2 $1 . $rest) '(declr-STAR))
+   (lambda ($3 $2 $1 . $rest) '(declr-star))
    ;; direct-abstract-declarator => direct-abstract-declarator "(" paramete...
    (lambda ($4 $3 $2 $1 . $rest)
-     `(declr-fctn ,$1 ,(tl->list $3)))
+     `(abs-ftn-declr ,$1 ,(tl->list $3)))
    ;; direct-abstract-declarator => direct-abstract-declarator "(" ")"
-   (lambda ($3 $2 $1 . $rest) `(declr-fctn ,$1))
+   (lambda ($3 $2 $1 . $rest) `(abs-ftn-declr ,$1))
    ;; direct-abstract-declarator => "(" parameter-type-list ")"
    (lambda ($3 $2 $1 . $rest)
-     `(declr-anon-fctn ,(tl->list $2)))
+     `(anon-ftn-declr ,(tl->list $2)))
    ;; direct-abstract-declarator => "(" ")"
-   (lambda ($2 $1 . $rest) '(declr-anon-fctn))
+   (lambda ($2 $1 . $rest) '(anon-ftn-declr))
    ;; typedef-name => 'typename
    (lambda ($1 . $rest) `(typename ,$1))
    ;; initializer => assignment-expression
@@ -622,9 +630,9 @@
    ;; designator-list => designator-list designator
    (lambda ($2 $1 . $rest) (tl-append $1 $2))
    ;; designator => "[" constant-expression "]"
-   (lambda ($3 $2 $1 . $rest) (list 'array-dsgr $2))
+   (lambda ($3 $2 $1 . $rest) `(array-dsgr ,$2))
    ;; designator => "." identifier
-   (lambda ($2 $1 . $rest) (list 'sel-dsgr $2))
+   (lambda ($2 $1 . $rest) `(sel-dsgr ,$2))
    ;; statement => labeled-statement
    (lambda ($1 . $rest) $1)
    ;; statement => compound-statement
@@ -637,6 +645,8 @@
    (lambda ($1 . $rest) $1)
    ;; statement => jump-statement
    (lambda ($1 . $rest) $1)
+   ;; statement => asm-statement
+   (lambda ($1 . $rest) $1)
    ;; statement => cpp-statement
    (lambda ($1 . $rest) $1)
    ;; labeled-statement => identifier ":" statement
@@ -702,6 +712,64 @@
    (lambda ($3 $2 $1 . $rest) `(return ,$2))
    ;; jump-statement => "return" ";"
    (lambda ($2 $1 . $rest) `(return (expr)))
+   ;; asm-statement => asm-expression ";"
+   (lambda ($2 $1 . $rest) $1)
+   ;; asm-expression => "asm" "(" string-literal ")"
+   (lambda ($4 $3 $2 $1 . $rest)
+     `(asm-expr (@ (extension "GNUC")) ,$3))
+   ;; asm-expression => "asm" "(" string-literal asm-outputs ")"
+   (lambda ($5 $4 $3 $2 $1 . $rest)
+     `(asm-expr
+        (@ (extension "GNUC"))
+        ,$3
+        ,(tl->list $4)))
+   ;; asm-expression => "asm" "(" string-literal asm-outputs asm-inputs ")"
+   (lambda ($6 $5 $4 $3 $2 $1 . $rest)
+     `(asm-expr
+        (@ (extension "GNUC"))
+        ,$3
+        ,(tl->list $4)
+        ,(tl->list $5)))
+   ;; asm-expression => "asm" "(" string-literal asm-outputs asm-inputs asm...
+   (lambda ($7 $6 $5 $4 $3 $2 $1 . $rest)
+     `(asm-expr
+        (@ (extension "GNUC"))
+        ,$3
+        ,(tl->list $4)
+        ,(tl->list $5)
+        ,(tl->list $6)))
+   ;; asm-outputs => ":"
+   (lambda ($1 . $rest) (make-tl 'asm-outputs))
+   ;; asm-outputs => ":" asm-output
+   (lambda ($2 $1 . $rest)
+     (make-tl 'asm-outputs $2))
+   ;; asm-outputs => asm-outputs "," asm-output
+   (lambda ($3 $2 $1 . $rest) (tl-append $1 $3))
+   ;; asm-output => string-literal "(" identifier ")"
+   (lambda ($4 $3 $2 $1 . $rest)
+     `(asm-operand ,$1 ,$3))
+   ;; asm-output => "[" identifier "]" string-literal "(" identifier ")"
+   (lambda ($7 $6 $5 $4 $3 $2 $1 . $rest)
+     `(asm-operand ,$2 ,$4 ,$6))
+   ;; asm-inputs => ":"
+   (lambda ($1 . $rest) (make-tl 'asm-inputs))
+   ;; asm-inputs => ":" asm-input
+   (lambda ($2 $1 . $rest) (make-tl 'asm-inputs $2))
+   ;; asm-inputs => asm-inputs "," asm-input
+   (lambda ($3 $2 $1 . $rest) (tl-append $1 $3))
+   ;; asm-input => string-literal "(" expression ")"
+   (lambda ($4 $3 $2 $1 . $rest)
+     `(asm-operand ,$1 ,$3))
+   ;; asm-input => "[" identifier "]" string-literal "(" expression ")"
+   (lambda ($7 $6 $5 $4 $3 $2 $1 . $rest)
+     `(asm-operand ,$2 ,$4 ,$6))
+   ;; asm-clobbers => ":"
+   (lambda ($1 . $rest) (make-tl 'asm-clobbers))
+   ;; asm-clobbers => ":" string-literal
+   (lambda ($2 $1 . $rest)
+     (make-tl 'asm-clobbers $2))
+   ;; asm-clobbers => asm-clobbers "," string-literal
+   (lambda ($3 $2 $1 . $rest) (tl-append $1 $3))
    ;; translation-unit => external-declaration-list
    (lambda ($1 . $rest) (tl->list $1))
    ;; external-declaration-list => 
@@ -729,7 +797,7 @@
         (extern-end)))
    ;; external-declaration => ";"
    (lambda ($1 . $rest)
-     `(decl (@ (extension . "GNU C"))))
+     `(decl (@ (extension "GNUC"))))
    ;; $P2 => 
    (lambda ($3 $2 $1 . $rest) (cpi-dec-blev!))
    ;; $P3 => 
@@ -772,8 +840,9 @@
    (lambda ($1 . $rest) `(comment ,$1))
    ;; cpp-statement => 'cpp-stmt
    (lambda ($1 . $rest) `(cpp-stmt ,$1))
-   ;; pragma => 'cpp-pragma
-   (lambda ($1 . $rest) $1)
+   ;; pragma => "_Pragma" "(" string-literal ")"
+   (lambda ($4 $3 $2 $1 . $rest)
+     `(pragma ,(tl->list $3)))
    ))
 
 ;;; end tables
